/*
 * Copyright (c) 2019-2020 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *	- Redistributions of source code must retain the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer.
 *
 *	- Redistributions in binary form must reproduce the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer in the documentation and/or other materials
 *	  provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#ifndef _AR_KDOR_ALGORITHM_H_
#define _AR_KDOR_ALGORITHM_H_

#include "opensm/osm_armgr_types.hpp"
#include "opensm/osm_ar_plft_based_algorithm.hpp"

#define MAX_KDOR_PLFT_NUMBER 4
#define MIN_KDOR_VL_CAP 3

typedef uint8_t port_index_t;

namespace OSM {

class KdorSwData;
class ArKdorAlgorithm;

struct KdorGroupData {
	GroupData* group_data_array[MAX_KDOR_PLFT_NUMBER - 1];
};

typedef map <u_int16_t, KdorGroupData> LidToKdorGroupDataMap;
typedef LidToKdorGroupDataMap::iterator LidToKdorGroupDataMapIter;
typedef LidToKdorGroupDataMap::const_iterator LidToKdorGroupDataMapConstIter;
typedef pair <LidToKdorGroupDataMapIter, bool> LidToKdorGroupDataMapInsertRes;

struct KdorAlgorithmData : public TreeAlgorithmData {

	LidToKdorGroupDataMap m_lid_to_kdor_group_map;

	/*
	 * Add lid to non empty group_data in ginen kdor_group_data
	 * And update lid_to_kdor_group_map
	 */
	void AddLidToKdorGroupData(uint16_t lid_num,
				   uint16_t sw_lid_num,
				   KdorGroupData &kdor_group_data,
				   bool is_new_group,
				   osm_log_t *p_osm_log);
private:

	void AddLidToKdorGroupData(uint16_t lid_num,
				   KdorGroupData &kdor_group_data);
};

class KdorSwData : public ArAlgorithmSwData {

public:
	KdorSwData(ARSWDataBaseEntry &db_entry,
		   OSMParallelPortGroupsCalculator *p_port_groups_calculator) :
		ArAlgorithmSwData(db_entry, p_port_groups_calculator) {}
};

class ArKdorAlgorithm : public PlftBasedArAlgorithm {

protected:
	KdorSwData     *m_sw_lid_to_kdor_data_[IB_LID_UCAST_END_HO + 1];

	sl2vl_t     m_turn_1_vl2vl_per_op_vls_[MAX_OP_VL_CODE+1];
	sl2vl_t     m_turn_2_vl2vl_per_op_vls_[MAX_OP_VL_CODE+1];

	ib_port_sl_to_plft_map_t m_plft_map_;

public:
	ArKdorAlgorithm(osm_log_t *p_osm_log,
			GuidToSWDataBaseEntry &sw_map,
			OSMAdaptiveRoutingManager &ar_mgr,
			supported_feature_t algorithm_feature,
			uint8_t planes_number,
			uint8_t min_vl_number) :
	PlftBasedArAlgorithm(p_osm_log, sw_map, ar_mgr,
			     algorithm_feature, planes_number, min_vl_number) {
		memset(m_sw_lid_to_kdor_data_, 0, sizeof(m_sw_lid_to_kdor_data_));
		BuildKdorVl2VlPerOpVl();
		BuildKdorPlftMap();
	}

	virtual ~ArKdorAlgorithm(){}

	virtual int Preprocess();
	virtual int RunCycle();

	virtual void Clear() {}
	virtual int Init();

	virtual void CalculateSwitchPortGroups(
		ARSWDataBaseEntry &sw_db_entry,
		LidMapping *p_lid_mapping,
		LidToConnectionArray &lid_to_connections,
		LidToPortArray &base_lid_to_port);

	virtual void CalculateRouteInfo(SwitchConnection &connection,
					RouteInfo &remote_route_info,
					RouteInfo &route_info);

	virtual int SetDirection(osm_switch_t* src,
				 SwitchConnection &connection) = 0;

protected:
	virtual ArAlgorithmSwData *GetArAlgorithmSwData(
		ARSWDataBaseEntry &db_entry) const { return db_entry.m_kdor_data; }

	virtual uint8_t GetMinimumVlCap() const { return MIN_KDOR_VL_CAP; }

	virtual void RestorePriv(ARSWDataBaseEntry &sw_db_entry);

	/*
	 * If flip_from_connection is true, the real from_connection is opposite
	 * to the given one and therefor has an opposite from dim_sign
	 */
	virtual ar_algorithm_turn_t GetTurnType(
		const SwitchConnection &from_connection,
		const SwitchConnection &to_connection,
		bool flip_from_connection) const;

private:
	bool SetCapable();

	/* Buils static unicast route info */
	int BuildStaticRouteInfo();

	/* Calculate Groups and pLFTs */
	void CalculateArGroups(ARSWDataBaseEntry &sw_db_entry,
			       u_int16_t *lid_to_sw_lid_mapping,
			       KdorAlgorithmData &algorithm_data);

	void CalculateArPlfts(ARSWDataBaseEntry &sw_db_entry,
			      LidMapping *p_lid_mapping,
			      KdorAlgorithmData &algorithm_data,
			      LidToConnectionArray &lid_to_connections,
			      LidToPortArray &base_lid_to_port);

	void CalculateArPlft(ARSWDataBaseEntry &sw_db_entry,
			     KdorAlgorithmData &algorithm_data,
			     u_int8_t ucast_lft_port,
			     unsigned dest_lid,
			     int plft_id,
			     LidToKdorGroupDataMapConstIter &ar_kdor_groups_iter,
			     OUT uint8_t &lid_state,
			     OUT uint16_t &ar_group_id);

	uint8_t GetStaticUcastLftPort(ARSWDataBaseEntry &sw_db_entry,
				      LidMapping *p_lid_mapping,
				      SwitchConnection *dst_sw_lid_to_conection[],
				      uint8_t base_lid_to_port[],
				      uint16_t dest_lid);

	/* Get static routing from the given switch */
	void BuildDstSwLidToConnection(ARSWDataBaseEntry &sw_db_entry,
				       SwitchConnection *dst_sw_lid_to_conection[]);

	void BuildKdorVl2VlPerOpVl();
	void BuildKdorPlftMap();

	void SetPlftMap();
	void SetPlftMap(ARSWDataBaseEntry &sw_db_entry);
};

#define HC_PLANES_NUMBER 4

class ArHcAlgorithm : public ArKdorAlgorithm {

public:

	ArHcAlgorithm(osm_log_t *p_osm_log,
		      GuidToSWDataBaseEntry &sw_map,
		      OSMAdaptiveRoutingManager &ar_mgr) :
	ArKdorAlgorithm(p_osm_log, sw_map, ar_mgr,
			SUPPORT_HC,
			HC_PLANES_NUMBER, HC_PLANES_NUMBER) {}

	virtual ~ArHcAlgorithm(){}

	virtual int SetDirection(osm_switch_t* src,
				 SwitchConnection &connection);

	virtual ar_algorithm_t GetAlgorithm() { return AR_ALGORITHM_KDOR_HC;}

};
} /* end of namespace OSM */

#endif /*_AR_KDOR_ALGORITHM_H_ */
